# -*- coding: utf-8 -*-

# Copyright 2015 Spanish National Research Council
# Copyright 2015 LIP - INDIGO-DataCloud
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import copy
import json
import os

import six.moves.urllib.parse as urlparse

from ooi import exception
from ooi.log import log as logging
from ooi.openstack import helpers as os_helpers
from ooi import utils

import webob.exc

LOG = logging.getLogger(__name__)


def _resolve_id(base_url, resource_url):
    """Gets the resource id from a base URL.

    :param base_url: application or request url (normally absolute)
    :param resource_url: absolute or relative resource url
    :returns: a tuple with the calculated base url and resource id
    """
    if not base_url.endswith('/'):
        base_url = base_url + '/'
    full_url = urlparse.urljoin(base_url, resource_url)
    parts = urlparse.urlsplit(full_url)
    base_parts = parts[0:2] + (os.path.dirname(parts[2]),) + parts[3:]
    return urlparse.urlunsplit(base_parts), os.path.basename(parts[2])


def get_id_with_kind(req, resource_url, kind=None):
    """Resolves the resource URL and tries to match it with the kind.

    :param req: current request
    :param resource_url: absolute or relative resource url
    :returns: a tuple with a base url and a resource id
    :raises ooi.exception.Invalid: if resource does not match kind
    """
    res_base, res_id = _resolve_id(req.url, resource_url)
    if kind:
        kind_base, kind_id = _resolve_id(req.application_url, kind.location)
        if kind_base != res_base:
            raise exception.Invalid("Expecting %s resource" % kind_base)
    return res_base, res_id


def exception_from_response(response):
    """Convert an OpenStack V2 Fault into a webob exception.

    Since we are calling the OpenStack API we should process the Faults
    produced by them. Extract the Fault information according to [1] and
    convert it back to a webob exception.

    [1] http://docs.openstack.org/developer/nova/v2/faults.html

    :param response: a webob.Response containing an exception
    :returns: a webob.exc.exception object
    """
    exceptions = {
        400: webob.exc.HTTPBadRequest,
        401: webob.exc.HTTPUnauthorized,
        403: webob.exc.HTTPForbidden,
        404: webob.exc.HTTPNotFound,
        405: webob.exc.HTTPMethodNotAllowed,
        406: webob.exc.HTTPNotAcceptable,
        409: webob.exc.HTTPConflict,
        413: webob.exc.HTTPRequestEntityTooLarge,
        415: webob.exc.HTTPUnsupportedMediaType,
        429: webob.exc.HTTPTooManyRequests,
        501: webob.exc.HTTPNotImplemented,
        503: webob.exc.HTTPServiceUnavailable,
    }
    code = response.status_int
    exc = exceptions.get(code, webob.exc.HTTPInternalServerError)
    try:
        message = response.json_body.popitem()[1].get("message")
        exc = exc(explanation=message)
    except Exception:
        LOG.exception("Unknown error happenened processing response %s"
                      % response)
        return webob.exc.HTTPInternalServerError()
    return exc


class BaseHelper(object):
    """Base helper to interact with nova API."""
    def __init__(self, app, openstack_version):
        self.app = app
        self.openstack_version = openstack_version

    def _get_req(self, req, method,
                 path=None,
                 content_type="application/json",
                 body=None,
                 query_string=""):
        """Return a new Request object to interact with OpenStack.

        This method will create a new request starting with the same WSGI
        environment as the original request, prepared to interact with
        OpenStack. Namely, it will override the script name to match the
        OpenStack version. It will also override the path, content_type and
        body of the request, if any of those keyword arguments are passed.

        :param req: the original request
        :param path: new path for the request
        :param content_type: new content type for the request, defaults to
                             "application/json" if not specified
        :param body: new body for the request
        :param query_string: query string for the request, defaults to an empty
                             query if not specified
        :returns: a Request object
        """
        if hasattr(self, 'neutron_endpoint'):
            server = self.neutron_endpoint
            environ = copy.copy(req.environ)
            try:
                if "HTTP_X-Auth-Token" not in environ:
                    env_token = req.environ["keystone.token_auth"]
                    token = env_token.get_auth_ref(None)['auth_token']
                    environ = {"HTTP_X-Auth-Token": token}
            except Exception:
                raise webob.exc.HTTPUnauthorized

            new_req = webob.Request.blank(path=path,
                                          environ=environ, base_url=server)
        else:
            new_req = webob.Request(copy.copy(req.environ))
            new_req.script_name = self.openstack_version
        new_req.query_string = query_string
        new_req.method = method
        if path is not None:
            new_req.path_info = path
        if content_type is not None:
            new_req.content_type = content_type
        if body is not None:
            new_req.body = utils.utf8(body)
        return new_req

    @staticmethod
    def get_from_response(response, element, default):
        """Get a JSON element from a valid response or raise an exception.

        This method will extract an element a JSON response (falling back to a
        default value) if the response has a code of 200, otherwise it will
        raise a webob.exc.exception

        :param response: The webob.Response object
        :param element: The element to look for in the JSON body
        :param default: The default element to be returned if not found.
        """
        if response.status_int in [200, 201, 202]:
            if element:
                return response.json_body.get(element, default)
            else:
                return response.json_body
        elif response.status_int in [204]:
            return []
        else:
            raise exception_from_response(response)

    def _make_get_request(self, req, path, parameters=None):
        """Create GET request

        This method creates a GET Request instance

        :param req: the incoming request
        :param path: element location
        :param parameters: parameters to filter results
        :param tenant: include tenant in the query parameters
        """
        query_string = utils.get_query_string(parameters)
        return self._get_req(req, path=path,
                             query_string=query_string, method="GET")

    def _make_create_request(self, req, resource, parameters):
        """Create CREATE request

        This method creates a CREATE Request instance

        :param req: the incoming request
        :param parameters: parameters with values
        """
        path = "/%s" % resource
        single_resource = resource[:-1]
        body = utils.make_body(single_resource, parameters)
        return self._get_req(req, path=path,
                             content_type="application/json",
                             body=json.dumps(body), method="POST")

    def _make_delete_request(self, req, path, id):
        """Create DELETE request

        This method creates a DELETE Request instance

        :param req: the incoming request
        :param path: element location
        """
        path = "%s/%s" % (path, id)
        return self._get_req(req, path=path, method="DELETE")

    def _make_put_request(self, req, path, parameters):
        """Create DELETE request

        This method creates a DELETE Request instance

        :param req: the incoming request
        :param path: element location
        """
        body = utils.make_body(None, parameters)
        return self._get_req(req, path=path,
                             content_type="application/json",
                             body=json.dumps(body), method="PUT")


class OpenStackHelper(BaseHelper):
    """Class to interact with the nova API."""
    required = {"networks": {"occi.core.title": "label",
                             "occi.network.address": "cidr",
                             }
                }

    @staticmethod
    def tenant_from_req(req):
        try:
            return req.environ["HTTP_X_PROJECT_ID"]
        except KeyError:
            raise exception.Forbidden(reason="Cannot find project ID")

    def _get_index_req(self, req):
        tenant_id = self.tenant_from_req(req)
        path = "/%s/servers" % tenant_id
        return self._get_req(req, path=path, method="GET")

    def index(self, req):
        """Get a list of servers for a tenant.

        :param req: the incoming request
        """
        os_req = self._get_index_req(req)
        response = os_req.get_response(self.app)
        return self.get_from_response(response, "servers", [])

    def _get_delete_req(self, req, server_id):
        tenant_id = self.tenant_from_req(req)
        path = "/%s/servers/%s" % (tenant_id, server_id)
        return self._get_req(req, path=path, method="DELETE")

    def delete(self, req, server_id):
        """Delete a server.

        :param req: the incoming request
        :param server_id: server id to delete
        """
        os_req = self._get_delete_req(req, server_id)
        response = os_req.get_response(self.app)
        # FIXME(aloga): this should be handled in get_from_response, shouldn't
        # it?
        if response.status_int not in [204]:
            raise exception_from_response(response)

    def _get_run_action_req(self, req, action, server_id):
        tenant_id = self.tenant_from_req(req)
        path = "/%s/servers/%s/action" % (tenant_id, server_id)

        actions_map = {
            "stop": {"os-stop": None},
            "start": {"os-start": None},
            "suspend": {"suspend": None},
            "resume": {"resume": None},
            "unpause": {"unpause": None},
            "restart": {"reboot": {"type": "SOFT"}},
        }
        action = actions_map[action]

        body = json.dumps(action)
        return self._get_req(req, path=path, body=body, method="POST")

    def run_action(self, req, action, server_id):
        """Run an action on a server.

        :param req: the incoming request
        :param action: the action to run
        :param server_id: server id to delete
        """
        os_req = self._get_run_action_req(req, action, server_id)
        response = os_req.get_response(self.app)
        if response.status_int != 202:
            raise exception_from_response(response)

    def _get_server_req(self, req, server_id):
        tenant_id = self.tenant_from_req(req)
        path = "/%s/servers/%s" % (tenant_id, server_id)
        return self._get_req(req, path=path, method="GET")

    def get_server(self, req, server_id):
        """Get info from a server.

        :param req: the incoming request
        :param server_id: server id to get info from
        """
        req = self._get_server_req(req, server_id)
        response = req.get_response(self.app)
        return self.get_from_response(response, "server", {})

    def _get_create_server_req(self, req, name, image, flavor,
                               user_data=None,
                               key_name=None,
                               block_device_mapping_v2=None,
                               networks=None):
        tenant_id = self.tenant_from_req(req)
        path = "/%s/servers" % tenant_id
        # TODO(enolfc): add here the correct metadata info
        # if contextualization.public_key.scheme in obj["schemes"]:
        #     req_body["metadata"] = XXX
        body = {"server": {
            "name": name,
            "imageRef": image,
            "flavorRef": flavor,
        }}
        if user_data is not None:
            body["server"]["user_data"] = user_data
        if key_name is not None:
            body["server"]["key_name"] = key_name
        if block_device_mapping_v2:
            body["server"]["block_device_mapping_v2"] = block_device_mapping_v2
        if networks:
            body["server"]["networks"] = networks

        return self._get_req(req,
                             path=path,
                             content_type="application/json",
                             body=json.dumps(body),
                             method="POST")

    def create_server(self, req, name, image, flavor,
                      user_data=None, key_name=None,
                      block_device_mapping_v2=None,
                      networks=None):
        """Create a server.

        :param req: the incoming request
        :param name: name for the new server
        :param image: image id for the new server
        :param flavor: flavor id for the new server
        :param user_data: user data to inject into the server
        :param key_name: user public key name
        """
        req = self._get_create_server_req(
            req,
            name,
            image,
            flavor,
            user_data=user_data,
            key_name=key_name,
            block_device_mapping_v2=block_device_mapping_v2,
            networks=networks)
        response = req.get_response(self.app)
        # We only get one server
        return self.get_from_response(response, "server", {})

    def _get_flavors_req(self, req):
        tenant_id = self.tenant_from_req(req)
        path = "/%s/flavors/detail" % tenant_id
        return self._get_req(req, path=path, method="GET")

    def get_flavors(self, req):
        """Get information from all flavors.

        :param req: the incoming request
        """
        req = self._get_flavors_req(req)
        response = req.get_response(self.app)
        return self.get_from_response(response, "flavors", [])

    def _get_flavor_req(self, req, flavor_id):
        tenant_id = self.tenant_from_req(req)
        path = "/%s/flavors/%s" % (tenant_id, flavor_id)
        return self._get_req(req, path=path, method="GET")

    def get_flavor(self, req, flavor_id):
        """Get information from a flavor.

        :param req: the incoming request
        :param flavor_id: flavor id to get info from
        """
        req = self._get_flavor_req(req, flavor_id)
        response = req.get_response(self.app)
        return self.get_from_response(response, "flavor", {})

    def _get_images_req(self, req):
        tenant_id = self.tenant_from_req(req)
        path = "/%s/images/detail" % tenant_id
        return self._get_req(req, path=path, method="GET")

    def get_images(self, req):
        """Get information from all images.

        :param req: the incoming request
        """
        req = self._get_images_req(req)
        response = req.get_response(self.app)
        return self.get_from_response(response, "images", [])

    def _get_image_req(self, req, image_id):
        tenant_id = self.tenant_from_req(req)
        path = "/%s/images/%s" % (tenant_id, image_id)
        return self._get_req(req, path=path, method="GET")

    def get_image(self, req, image_id):
        """Get information from an image.

        :param req: the incoming request
        :param image_id: image id to get info from
        """
        req = self._get_image_req(req, image_id)
        response = req.get_response(self.app)
        return self.get_from_response(response, "image", {})

    def _get_volumes_req(self, req):
        tenant_id = self.tenant_from_req(req)
        path = "/%s/os-volumes" % tenant_id
        return self._get_req(req, path=path, method="GET")

    def get_volumes(self, req):
        req = self._get_volumes_req(req)
        response = req.get_response(self.app)
        return self.get_from_response(response, "volumes", [])

    def _get_volume_req(self, req, volume_id):
        tenant_id = self.tenant_from_req(req)
        path = "/%s/os-volumes/%s" % (tenant_id, volume_id)
        return self._get_req(req, path=path, method="GET")

    def get_volume(self, req, volume_id):
        """Get information from a volume.

        :param req: the incoming request
        :param volume_id: volume id to get info from
        """
        req = self._get_volume_req(req, volume_id)
        response = req.get_response(self.app)
        return self.get_from_response(response, "volume", [])

    def _get_server_volumes_link_req(self, req, server_id):
        tenant_id = self.tenant_from_req(req)
        path = "/%s/servers/%s/os-volume_attachments" % (tenant_id, server_id)
        return self._get_req(req, path=path, method="GET")

    def get_server_volumes_link(self, req, server_id):
        """Get volumes attached to a server.

        :param req: the incoming request
        :param server_id: server id to get volumes from
        """
        req = self._get_server_volumes_link_req(req, server_id)
        response = req.get_response(self.app)
        return self.get_from_response(response, "volumeAttachments", [])

    def _get_server_volumes_link_create_req(self, req, s_id, v_id, dev=None):
        tenant_id = self.tenant_from_req(req)
        body = {
            "volumeAttachment": {
                "volumeId": v_id
            }
        }
        if dev is not None:
            body["volumeAttachment"]["device"] = dev
        path = "/%s/servers/%s/os-volume_attachments" % (tenant_id, s_id)
        return self._get_req(req,
                             path=path,
                             content_type="application/json",
                             body=json.dumps(body),
                             method="POST")

    def create_server_volumes_link(self, req, server_id, vol_id, dev=None):
        req = self._get_server_volumes_link_create_req(req, server_id, vol_id,
                                                       dev=dev)
        response = req.get_response(self.app)
        return self.get_from_response(response, "volumeAttachment", {})

    def _get_server_volumes_link_delete_req(self, req, server_id, vol_id):
        tenant_id = self.tenant_from_req(req)
        path = "/%s/servers/%s/os-volume_attachments/%s" % (tenant_id,
                                                            server_id, vol_id)
        return self._get_req(req, path=path, method="DELETE")

    def delete_server_volumes_link(self, req, server_id, vol_id):
        req = self._get_server_volumes_link_delete_req(req, server_id, vol_id)
        response = req.get_response(self.app)
        if response.status_int not in [202]:
            raise exception_from_response(response)

    def _get_floating_ips_req(self, req):
        tenant_id = self.tenant_from_req(req)
        path = "/%s/os-floating-ips" % tenant_id
        return self._get_req(req, path=path, method="GET")

    def get_floating_ips(self, req):
        """Get floating IPs for the tenant.

        :param req: the incoming request
        """
        req = self._get_floating_ips_req(req)
        response = req.get_response(self.app)
        return self.get_from_response(response, "floating_ips", [])

    def _get_floating_ip_pools_req(self, req):
        tenant_id = self.tenant_from_req(req)
        path = "/%s/os-floating-ip-pools" % tenant_id
        return self._get_req(req, path=path, method="GET")

    def get_floating_ip_pools(self, req):
        """Get floating IP pools for the tenant.

        :param req: the incoming request
        """
        req = self._get_floating_ip_pools_req(req)
        response = req.get_response(self.app)
        return self.get_from_response(response, "floating_ip_pools", [])

    def _get_volume_delete_req(self, req, vol_id):
        tenant_id = self.tenant_from_req(req)
        path = "/%s/os-volumes/%s" % (tenant_id, vol_id)
        return self._get_req(req, path=path, method="DELETE")

    def volume_delete(self, req, vol_id):
        """Delete a volume.

        :param req: the incoming request
        :param vol_id: volume id to delete
        """
        req = self._get_volume_delete_req(req, vol_id)
        response = req.get_response(self.app)
        # FIXME(aloga): this should be handled in get_from_response, shouldn't
        # it?
        if response.status_int not in [204]:
            raise exception_from_response(response)

    def _get_volume_create_req(self, req, name, size):
        tenant_id = self.tenant_from_req(req)
        path = "/%s/os-volumes" % tenant_id
        body = {"volume": {
            "display_name": name,
            "size": size,
        }}
        return self._get_req(req,
                             path=path,
                             content_type="application/json",
                             body=json.dumps(body),
                             method="POST")

    def volume_create(self, req, name, size):
        """Create a volume.

        :param req: the incoming request
        :param name: name for the new volume
        :param size: size for the new volume
        """
        req = self._get_volume_create_req(req, name, size)
        response = req.get_response(self.app)
        # We only get one volume
        return self.get_from_response(response, "volume", {})

    def _get_floating_ip_allocate_req(self, req, pool=None):
        tenant_id = self.tenant_from_req(req)
        path = "/%s/os-floating-ips" % tenant_id
        body = {"pool": pool}
        return self._get_req(req, path=path, body=json.dumps(body),
                             method="POST")

    def allocate_floating_ip(self, req, pool=None):
        """Allocate a floating ip from a pool.

        :param req: the incoming request
        :param pool: floating ip pool to get the IP from
        """
        req = self._get_floating_ip_allocate_req(req, pool)
        response = req.get_response(self.app)
        return self.get_from_response(response, "floating_ip", {})

    def _get_floating_ip_release_req(self, req, ip):
        tenant_id = self.tenant_from_req(req)
        path = "/%s/os-floating-ips/%s" % (tenant_id, ip)
        return self._get_req(req, path=path, method="DELETE")

    def release_floating_ip(self, req, ip):
        """Release a floating ip.

        :param req: the incoming request
        :param ip: floating ip pool to release
        """
        req = self._get_floating_ip_release_req(req, ip)
        response = req.get_response(self.app)
        if response.status_int != 202:
            raise exception_from_response(response)

    def _get_associate_floating_ip_req(self, req, server, address):
        tenant_id = self.tenant_from_req(req)
        body = {"addFloatingIp": {"address": address}}
        path = "/%s/servers/%s/action" % (tenant_id, server)
        return self._get_req(req, path=path, body=json.dumps(body),
                             method="POST")

    def associate_floating_ip(self, req, server, address):
        """Associate a floating ip to a server.

        :param req: the incoming request
        :param server: the server to associate the ip to
        :param address: ip to associate to the server
        """
        req = self._get_associate_floating_ip_req(req, server, address)
        response = req.get_response(self.app)
        if response.status_int != 202:
            raise exception_from_response(response)

    def _get_remove_floating_ip_req(self, req, server, address):
        tenant_id = self.tenant_from_req(req)
        body = {"removeFloatingIp": {"address": address}}
        path = "/%s/servers/%s/action" % (tenant_id, server)
        return self._get_req(req, path=path, body=json.dumps(body),
                             method="POST")

    def remove_floating_ip(self, req, server, address):
        """Remove a floating ip to a server.

        :param req: the incoming request
        :param server: the server to remove the ip from
        :param address: ip to remove from the server
        """
        req = self._get_remove_floating_ip_req(req, server, address)
        response = req.get_response(self.app)
        if response.status_int != 202:
            raise exception_from_response(response)

    def _get_keypair_create_req(self, req, name, public_key=None):
        tenant_id = self.tenant_from_req(req)
        path = "/%s/os-keypairs" % tenant_id
        body = {"keypair": {
            "name": name,
        }}
        if public_key:
            body["keypair"]["public_key"] = public_key

        return self._get_req(req,
                             path=path,
                             content_type="application/json",
                             body=json.dumps(body),
                             method="POST")

    def keypair_create(self, req, name, public_key=None):
        """Create a keypair.

        :param req: the incoming request
        :param name: name for the new keypair
        :param public_key: public ssh key to import
        """
        req = self._get_keypair_create_req(req, name, public_key=public_key)
        response = req.get_response(self.app)
        return self.get_from_response(response, "keypair", {})

    def _get_keypair_delete_req(self, req, name):
        tenant_id = self.tenant_from_req(req)
        path = "/%s/os-keypairs/%s" % (tenant_id, name)
        return self._get_req(req, path=path, method="DELETE")

    def keypair_delete(self, req, name):
        """Delete a keypair.

        :param req: the incoming request
        :param name: keypair name to delete
        """
        req = self._get_keypair_delete_req(req, name)
        response = req.get_response(self.app)
        # FIXME(orviz) API says 204 is the normal response code but it
        # is actually returning 202 (bug likely)
        if response.status_int not in [202, 204]:
            raise exception_from_response(response)

    @staticmethod
    def _build_link(net_id, compute_id, ip, ip_id=None, mac=None, pool=None,
                    state='ACTIVE'):
        link = {}
        link['mac'] = mac
        link['pool'] = pool
        link['network_id'] = net_id
        link['compute_id'] = compute_id
        link['ip'] = ip
        link['ip_id'] = ip_id
        link['state'] = os_helpers.vm_state(state)
        return link

    def _get_ports(self, req, compute_id):
        result = []
        tenant_id = self.tenant_from_req(req)
        path = "/%s/servers/%s/os-interface" % (tenant_id, compute_id)
        os_req = self._get_req(req, path=path, method="GET")
        response = os_req.get_response(self.app)
        try:
            result = self.get_from_response(response,
                                            "interfaceAttachments", [])
        except Exception as e:
            LOG.exception("Interfaces can not be shown: " + e.explanation)
        return result

    def get_compute_net_link(self, req, compute_id, network_id,
                             address):
        """Get a specific network/server link

        It shows a specific link (either private or public ip)

        :param req: the incoming request
        :param compute_id: server id
        :param network_id: network id
        :param address: ip connected
        """

        s = self.get_server(req, compute_id)
        pool = None
        ip_id = None
        server_addrs = s.get("addresses", {})
        for addr_set in server_addrs.values():
            for addr in addr_set:
                if addr["addr"] == address:
                    mac = addr["OS-EXT-IPS-MAC:mac_addr"]
                    if network_id == os_helpers.PUBLIC_NETWORK:
                        floating_ips = self.get_floating_ips(
                            req
                        )
                        for ip in floating_ips:
                            if address == ip['ip']:
                                pool = ip['pool']
                                ip_id = ip['id']
                    else:
                        ports = self._get_ports(req, compute_id)
                        for p in ports:
                            if p["net_id"] == network_id:
                                for ip in p["fixed_ips"]:
                                    if ip['ip_address'] == address:
                                        ip_id = p['port_id']
                    return self._build_link(network_id,
                                            compute_id,
                                            address,
                                            mac=mac,
                                            pool=pool,
                                            ip_id=ip_id
                                            )
        raise exception.NotFound()

    def list_compute_net_links(self, req):
        """Get floating IPs for the tenant.

        :param req: the incoming request
        """
        floating_ips = self.get_floating_ips(req)
        float_list = {}
        for ip in floating_ips:
            if ip["instance_id"]:
                float_list.update({ip['fixed_ip']: ip})
        servers = self.index(req)
        link_list = []
        for s in servers:
            ports = self._get_ports(req, s['id'])
            for p in ports:
                for ip in p["fixed_ips"]:
                    mac = p['mac_addr']
                    state = p["port_state"]
                    link = self._build_link(p["net_id"],
                                            s['id'],
                                            ip['ip_address'],
                                            ip_id=p["port_id"],
                                            mac=mac,
                                            state=state)
                    link_list.append(link)
                    float_ip = float_list.get(ip['ip_address'], None)
                    if float_ip:
                        link = self._build_link(p["net_id"],
                                                float_ip['instance_id'],
                                                float_ip['ip'],
                                                ip_id=float_ip["id"],
                                                mac=mac,
                                                pool=float_ip["pool"]
                                                )
                        link_list.append(link)
        if not link_list:
            for ip in floating_ips:
                if ip["instance_id"]:
                    link = self._build_link(os_helpers.PUBLIC_NETWORK,
                                            ip['instance_id'],
                                            ip['ip'],
                                            ip_id=ip["id"],
                                            pool=ip["pool"]
                                            )
                    link_list.append(link)
        return link_list

    def create_port(self, req, network_id, device_id):
        """Add a port to the subnet

        Returns the port information

        :param req: the incoming network
        :param network_id: network id
        :param device_id: device id
        """
        param_port = {
            'net_id': network_id
        }
        tenant_id = self.tenant_from_req(req)
        path = "/%s/servers/%s/os-interface" % (tenant_id, device_id)
        body = utils.make_body("interfaceAttachment", param_port)
        os_req = self._get_req(req, path=path,
                               content_type="application/json",
                               body=json.dumps(body), method="POST")
        response = os_req.get_response(self.app)
        port = self.get_from_response(response, "interfaceAttachment", {})
        for ip in port["fixed_ips"]:
            return self._build_link(port["net_id"],
                                    device_id,
                                    ip['ip_address'],
                                    ip_id=port["port_id"],
                                    mac=port['mac_addr'],
                                    state=port["port_state"])

    def delete_port(self, req, compute_id, ip_id):
        """Delete a port to the subnet

        Returns the port information

        :param req: the incoming network
        :param compute_id: compute id
        :param ip_id: ip id
        """
        path = "servers/%s/os-interface" % compute_id
        tenant_id = self.tenant_from_req(req)
        path = "/%s/%s/%s" % (tenant_id, path, ip_id)
        os_req = self._get_req(req, path=path, method="DELETE")
        os_req.get_response(self.app)
        return []

        raise exception.LinkNotFound(
            "Interface %s not found" % ip_id
        )

    def get_network_id(self, req, mac, server_id):
        """Get the Network ID from the mac port

        :param req: the incoming network
        :param mac: mac port
        :param server_id: server id
        """
        ports = self._get_ports(req, server_id)
        for p in ports:
            server_mac = p['mac_addr']
            if server_mac == mac:
                return p['net_id']

        raise webob.exc.HTTPNotFound

    def assign_floating_ip(self, req, network_id, device_id,
                           pool=None):
        """assign floating ip to a server

        :param req: the incoming request
        :param network_id: network id
        :param device_id: device id
        """
        # net_id it is not needed if
        # there is just one port of the VM
        # FIXME(jorgesece): raise an error if the first port has
        # already a floating-ip
        ip = self.allocate_floating_ip(req, pool)
        # Add it to server
        self.associate_floating_ip(req, device_id, ip["ip"])

        try:
            link_public = self._build_link(
                network_id,
                device_id,
                ip["ip"],
                ip_id=ip["id"],
                pool=ip["pool"])
        except Exception:
            raise exception.OCCIInvalidSchema()
        return link_public

    @staticmethod
    def _build_networks(networks):
        ooi_net_list = []
        for net in networks:
            # TODO(jorgesece): manage IP_v6
            ooi_net = {}
            ooi_net["address"] = net.get("cidr", None)
            ooi_net["state"] = "active"
            ooi_net["id"] = net["id"]
            ooi_net["name"] = net.get("label", None)
            ooi_net["gateway"] = net.get("gateway", None)
            ooi_net_list.append(ooi_net)
        return ooi_net_list

    def list_networks(self, req):
        """Get a list of servers for a tenant.

        :param req: the incoming request
        :param parameters: parameters with tenant
        """
        path = "os-networks"
        tenant_id = self.tenant_from_req(req)
        path = "/%s/%s" % (tenant_id, path)
        os_req = self._get_req(req, path=path,
                               method="GET")
        response = os_req.get_response(self.app)
        nets = self.get_from_response(response, "networks", [])
        ooi_networks = self._build_networks(nets)
        pools = self.get_floating_ip_pools(req)
        if pools:
            net = {'id': os_helpers.PUBLIC_NETWORK,
                   'label': os_helpers.PUBLIC_NETWORK}
            public_net = self._build_networks([net])[0]
            ooi_networks.append(public_net)
        return ooi_networks

    def get_network_details(self, req, id):
        """Get info from a network.

        It returns json code from the server

        :param req: the incoming network
        :param id: net identification
        """
        if id == os_helpers.PUBLIC_NETWORK:
            net = {'id': os_helpers.PUBLIC_NETWORK,
                   'label': 'PUBLIC_to_associate_Floating_IPs'}
        else:
            path = "os-networks/%s" % id
            tenant_id = self.tenant_from_req(req)
            path = "/%s/%s" % (tenant_id, path)
            os_req = self._get_req(req, path=path,
                                   method="GET")
            response = os_req.get_response(self.app)
            net = self.get_from_response(response, "network", {})
        ooi_networks = self._build_networks([net])
        return ooi_networks[0]

    def create_network(self, req, name, cidr,
                       gateway=None, ip_version=None):
        """Create a network in nova-network.

        :param req: the incoming request
        :param name: network resource to manage
        :param cidr: parameters with values
        :param gateway: gateway ip
        :param ip_version: ip version
        """
        net_param = {'label': name,
                     'cidr': cidr,
                     'gateway': gateway
                     }
        path = "os-networks"
        tenant_id = self.tenant_from_req(req)
        path = "/%s/%s" % (tenant_id, path)
        body = utils.make_body('network', net_param)
        os_req = self._get_req(req, path=path,
                               content_type="application/json",
                               body=json.dumps(body), method="POST")
        response = os_req.get_response(self.app)
        net = self.get_from_response(
            response, "network", {})
        ooi_net = self._build_networks([net])
        return ooi_net[0]

    def delete_network(self, req, id):
        """Delete a network.

        It returns json code from the server

        :param req: the incoming network
        :param id: net identification
        :param parameters: parameters with tenant
        """
        path = "os-networks"
        tenant_id = self.tenant_from_req(req)
        path = "/%s/%s/%s" % (tenant_id, path, id)
        os_req = self._get_req(req, path=path, method="DELETE")
        os_req.get_response(self.app)
        return []



# -*- coding: utf-8 -*-

# Copyright 2015 Spanish National Research Council
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import uuid

import mock

from ooi.api import helpers
from ooi.api import storage_link as storage_link_api
from ooi import exception
from ooi.occi.core import collection
from ooi.occi.infrastructure import compute
from ooi.occi.infrastructure import storage
from ooi.occi.infrastructure import storage_link
from ooi.tests import base
from ooi.tests import fakes


class TestStorageLinkController(base.TestController):
    def setUp(self):
        super(TestStorageLinkController, self).setUp()
        self.controller = storage_link_api.Controller(mock.MagicMock(), None)

    @mock.patch.object(helpers.OpenStackHelper, "get_volumes")
    def test_index(self, m_volumes):
        for tenant in fakes.tenants.values():
            vols = fakes.volumes[tenant["id"]]
            m_volumes.return_value = vols
            ret = self.controller.index(None)
            self.assertIsInstance(ret, collection.Collection)
            # NOTE(aloga): the only tenant with an attachment
            if tenant["name"] == "baz":
                for idx, vol in enumerate(vols):
                    self.assertIsInstance(ret.resources[idx],
                                          storage_link.StorageLink)
                    server_id = vol["attachments"][0]["server_id"]
                    link_id = "%s_%s" % (server_id, vol["id"])
                    self.assertEqual(link_id, ret.resources[idx].id)
            else:
                self.assertEqual([], ret.resources)
            m_volumes.assert_called_with(None)

    @mock.patch.object(helpers.OpenStackHelper, "delete_server_volumes_link")
    @mock.patch.object(storage_link_api.Controller, "_get_attachment_from_id")
    def test_delete(self, mock_get, mock_delete):
        server_id = uuid.uuid4().hex
        vol_id = uuid.uuid4().hex
        link_id = "%s_%s" % (server_id, vol_id)
        mock_get.return_value = {"serverId": server_id, "volumeId": vol_id}
        mock_delete.return_value = []
        ret = self.controller.delete(None, link_id)
        self.assertEqual([], ret)
        mock_delete.assert_called_with(None, server_id, vol_id)
        mock_get.assert_called_with(None, link_id)

    @mock.patch.object(storage_link_api.Controller, "_get_attachment_from_id")
    def test_show(self, m_get):
        server_id = uuid.uuid4().hex
        vol_id = uuid.uuid4().hex
        link_id = "%s_%s" % (server_id, vol_id)
        m_get.return_value = {
            "serverId": server_id,
            "volumeId": vol_id,
            "device": "/dev/sda",
        }
        ret = self.controller.show(None, link_id)
        link = ret.pop()
        self.assertIsInstance(link, storage_link.StorageLink)
        self.assertIsInstance(link.source, compute.ComputeResource)
        self.assertIsInstance(link.target, storage.StorageResource)
        self.assertEqual(vol_id, link.target.id)
        self.assertEqual(server_id, link.source.id)
        m_get.assert_called_with(None, link_id)

    @mock.patch.object(helpers.OpenStackHelper, "get_server_volumes_link")
    def test_get_attachment_from_id(self, m_get):
        server_uuid = uuid.uuid4().hex
        vol_uuid = uuid.uuid4().hex
        link_id = "%s_%s" % (server_uuid, vol_uuid)
        m_get.return_value = [
            {"volumeId": uuid.uuid4().hex},
            {"volumeId": vol_uuid},
        ]
        self.assertEqual({"volumeId": vol_uuid},
                         self.controller._get_attachment_from_id(None,
                                                                 link_id))
        link_id = "%s_%s" % (uuid.uuid4().hex, uuid.uuid4().hex)
        self.assertRaises(exception.LinkNotFound,
                          self.controller._get_attachment_from_id,
                          None,
                          link_id)

    def test_get_attachment_from_id_invalid(self):
        self.assertRaises(exception.LinkNotFound,
                          self.controller._get_attachment_from_id,
                          None,
                          "foobarbaz")

    @mock.patch.object(helpers.OpenStackHelper, "create_server_volumes_link")
    @mock.patch("ooi.occi.validator.Validator")
    @mock.patch("ooi.api.helpers.get_id_with_kind")
    def test_create_link(self, m_get_id, m_validator, m_create):
        tenant = fakes.tenants["foo"]
        req = self._build_req(tenant["id"])
        vol_id = uuid.uuid4().hex
        server_id = uuid.uuid4().hex
        obj = {
            "attributes": {
                "occi.core.target": vol_id,
                "occi.core.source": server_id,
            }
        }
        # NOTE(aloga): the mocked call is
        # "parser = req.get_parser()(req.headers, req.body)"
        req.get_parser = mock.MagicMock()
        # NOTE(aloga): MOG!
        req.get_parser.return_value.return_value.parse.return_value = obj
        m_validator.validate.return_value = True
        m_get_id.side_effect = [('', vol_id), ('', server_id)]
        attachment = {
            "device": "/dev/vdd",
            "id": "a26887c6-c47b-4654-abb5-dfadf7d3f803",
            "serverId": server_id,
            "volumeId": vol_id,
        }
        m_create.return_value = attachment
        ret = self.controller.create(req, None)
        link = ret.resources.pop()
        self.assertIsInstance(link, storage_link.StorageLink)
        self.assertIsInstance(link.source, compute.ComputeResource)
        self.assertIsInstance(link.target, storage.StorageResource)
        self.assertEqual(vol_id, link.target.id)
        self.assertEqual(server_id, link.source.id)
        m_create.assert_called_with(mock.ANY, server_id, vol_id, dev=None)



# -*- coding: utf-8 -*-

# Copyright 2015 Spanish National Research Council
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import uuid

import mock

from ooi.tests import fakes
from ooi.tests.functional.middleware import test_middleware
from ooi import utils


class TestStorageLinkController(test_middleware.TestMiddleware):
    """Test OCCI storage link controller."""
    def test_list_vols_empty(self):
        tenant = fakes.tenants["bar"]
        app = self.get_app()

        for url in ("/storagelink/", "/storagelink"):
            req = self._build_req(url, tenant["id"], method="GET")

            req.environ["HTTP_X_PROJECT_ID"] = tenant["id"]

            resp = req.get_response(app)

            expected_result = ""
            self.assertContentType(resp)
            self.assertExpectedResult(expected_result, resp)
            self.assertEqual(204, resp.status_code)

    def test_list_attachments_empty(self):
        tenant = fakes.tenants["foo"]
        app = self.get_app()

        for url in ("/storagelink/", "/storagelink"):
            req = self._build_req(url, tenant["id"], method="GET")

            m = mock.MagicMock()
            m.user.project_id = tenant["id"]
            req.environ["keystone.token_auth"] = m

            resp = req.get_response(app)

            expected_result = ""
            self.assertContentType(resp)
            self.assertExpectedResult(expected_result, resp)
            self.assertEqual(204, resp.status_code)

    def test_list_attachments(self):
        tenant = fakes.tenants["baz"]
        app = self.get_app()

        for url in ("/storagelink/", "/storagelink"):
            req = self._build_req(url, tenant["id"], method="GET")

            resp = req.get_response(app)

            self.assertEqual(200, resp.status_code)
            expected = []
            for v in fakes.volumes[tenant["id"]]:
                for a in v["attachments"]:
                    link_id = '_'.join([a["serverId"], v["id"]])
                    expected.append(
                        ("X-OCCI-Location",
                         utils.join_url(self.application_url + "/",
                                        "storagelink/%s" % link_id))
                    )
            self.assertExpectedResult(expected, resp)

    def test_show_link(self):
        tenant = fakes.tenants["baz"]
        app = self.get_app()

        for volume in fakes.volumes[tenant["id"]]:
            for a in volume["attachments"]:
                link_id = '_'.join([a["serverId"], volume["id"]])
                req = self._build_req("/storagelink/%s" % link_id,
                                      tenant["id"], method="GET")

                resp = req.get_response(app)
                self.assertContentType(resp)
                link_id = '_'.join([a["serverId"], a["volumeId"]])
                source = utils.join_url(self.application_url + "/",
                                        "compute/%s" % a["serverId"])
                target = utils.join_url(self.application_url + "/",
                                        "storage/%s" % a["volumeId"])
                self.assertResultIncludesLinkAttr(link_id, source, target,
                                                  resp)
                self.assertEqual(200, resp.status_code)

    def test_show_invalid_id(self):
        tenant = fakes.tenants["foo"]

        app = self.get_app()
        req = self._build_req("/storagelink/%s" % uuid.uuid4().hex,
                              tenant["id"], method="GET")
        resp = req.get_response(app)
        self.assertEqual(404, resp.status_code)

    def test_show_non_existant_compute(self):
        tenant = fakes.tenants["foo"]

        app = self.get_app()
        req = self._build_req("/storagelink/%s_foo" % uuid.uuid4().hex,
                              tenant["id"], method="GET")
        resp = req.get_response(app)
        self.assertEqual(404, resp.status_code)

    def test_show_non_existant_volume(self):
        tenant = fakes.tenants["foo"]
        server_id = fakes.servers[tenant["id"]][0]["id"]

        app = self.get_app()
        req = self._build_req("/storagelink/%s_foo" % server_id,
                              tenant["id"], method="GET")
        resp = req.get_response(app)
        self.assertEqual(404, resp.status_code)

    def test_create_link(self):
        tenant = fakes.tenants["foo"]

        server_id = fakes.servers[tenant["id"]][0]["id"]
        server_url = utils.join_url(self.application_url + "/",
                                    "compute/%s" % server_id)
        vol_id = fakes.volumes[tenant["id"]][0]["id"]
        vol_url = utils.join_url(self.application_url + "/",
                                 "storage/%s" % vol_id)

        app = self.get_app()
        headers = {
            'Category': (
                'storagelink;'
                'scheme="http://schemas.ogf.org/occi/infrastructure#";'
                'class="kind"'),
            'X-OCCI-Attribute': (
                'occi.core.source="%s", '
                'occi.core.target="%s"'
                ) % (server_url, vol_url)
        }
        req = self._build_req("/storagelink", tenant["id"], method="POST",
                              headers=headers)
        resp = req.get_response(app)

        link_id = '_'.join([server_id, vol_id])
        expected = [("X-OCCI-Location",
                     utils.join_url(self.application_url + "/",
                                    "storagelink/%s" % link_id))]
        self.assertEqual(200, resp.status_code)
        self.assertExpectedResult(expected, resp)
        self.assertDefaults(resp)

    def test_create_link_with_device(self):
        tenant = fakes.tenants["foo"]

        server_id = fakes.servers[tenant["id"]][0]["id"]
        server_url = utils.join_url(self.application_url + "/",
                                    "compute/%s" % server_id)
        vol_id = fakes.volumes[tenant["id"]][0]["id"]
        vol_url = utils.join_url(self.application_url + "/",
                                 "storage/%s" % vol_id)

        app = self.get_app()
        headers = {
            'Category': (
                'storagelink;'
                'scheme="http://schemas.ogf.org/occi/infrastructure#";'
                'class="kind"'),
            'X-OCCI-Attribute': (
                'occi.storagelink.deviceid="/dev/vdc", '
                'occi.core.source="%s", '
                'occi.core.target="%s"'
                ) % (server_url, vol_url)
        }
        req = self._build_req("/storagelink", tenant["id"], method="POST",
                              headers=headers)
        resp = req.get_response(app)

        link_id = '_'.join([server_id, vol_id])
        expected = [("X-OCCI-Location",
                     utils.join_url(self.application_url + "/",
                                    "storagelink/%s" % link_id))]
        self.assertEqual(200, resp.status_code)
        self.assertExpectedResult(expected, resp)
        self.assertDefaults(resp)

    def test_delete_link(self):
        tenant = fakes.tenants["baz"]
        app = self.get_app()

        for volume in fakes.volumes[tenant["id"]]:
            for a in volume["attachments"]:
                link_id = '_'.join([a["serverId"], volume["id"]])
                req = self._build_req("/storagelink/%s" % link_id,
                                      tenant["id"], method="DELETE")
                resp = req.get_response(app)
                self.assertContentType(resp)
                self.assertEqual(204, resp.status_code)


class StorageLinkControllerTextPlain(test_middleware.TestMiddlewareTextPlain,
                                     TestStorageLinkController):
    """Test OCCI storage link controller with Accept: text/plain."""


class StorageLinkControllerTextOcci(test_middleware.TestMiddlewareTextOcci,
                                    TestStorageLinkController):
    """Test OCCI storage link controller with Accept: text/occi."""



# -*- coding: utf-8 -*-

# Copyright 2015 Spanish National Research Council
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import copy

from ooi import exception
from ooi.occi import helpers


class Validator(object):
    def __init__(self, obj):
        self.parsed_obj = obj

    def _validate_category(self, category):
        try:
            if category.type_id != self.parsed_obj["category"]:
                raise exception.OCCISchemaMismatch(
                    expected=category.type_id,
                    found=self.parsed_obj["category"]
                )
        except KeyError:
            raise exception.OCCIMissingType(
                type_id=category.type_id)

    def _compare_schemes(self, expected_type, actual):
        actual_scheme, actual_term = helpers.decompose_type(actual)
        if expected_type.scheme != actual_scheme:
            return False
        try:
            if expected_type.term != actual_term:
                return False
        except AttributeError:
            # ignore the fact the type does not have a term
            pass
        return True

    def _validate_mandatory_mixins(self, mixins, unmatched):
        for m in mixins:
            for um in unmatched:
                if self._compare_schemes(m, um):
                    unmatched[um] -= 1
                    break
            else:
                # NOTE(aloga): I am not sure of this...
                expected = m.scheme
                if hasattr(m, "term"):
                    expected += m.term
                raise exception.OCCIMissingType(type_id=expected)
        return unmatched

    def _validate_optional_mixins(self, mixins, unmatched):
        for m in mixins:
            for um in unmatched:
                if self._compare_schemes(m, um):
                    unmatched[um] -= 1
                    break
        return unmatched

    def _validate_optional_links(self, expected, links):
        for uri, l in links.items():
            try:
                rel = l['rel']
            except KeyError:
                raise exception.OCCIMissingType(type_id=uri)
            for ex in expected:
                if rel == ex.type_id:
                    break
            else:
                expected_types = ', '.join([e.type_id for e in expected])
                raise exception.OCCISchemaMismatch(expected=expected_types,
                                                   found=l['rel'])

    def validate_attributes(self, required):
        """Validate required attributes

        :param required: required attributes
        """
        attr = self.parsed_obj.get("attributes", {})
        if required:
            for at in required:
                if at not in attr:
                    raise exception.Invalid(
                        "Expecting %s attribute" % at
                    )

    def validate(self, schema):
        if "category" in schema:
            self._validate_category(schema["category"])
        unmatched = copy.copy(self.parsed_obj["mixins"])
        unmatched = self._validate_mandatory_mixins(
            schema.get("mixins", []), unmatched)
        unmatched = self._validate_optional_mixins(
            schema.get("optional_mixins", []), unmatched)
        unexpected = [m for m in unmatched if unmatched[m]]
        if unexpected:
            raise exception.OCCISchemaMismatch(expected="",
                                               found=unexpected)
        self._validate_optional_links(
            schema.get("optional_links", []),
            self.parsed_obj.get("links", {}))
        return True



# -*- coding: utf-8 -*-

# Copyright 2015 Spanish National Research Council
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import mock

from ooi.api import helpers
from ooi.api import query
from ooi.occi.core import entity
from ooi.occi.core import link
from ooi.occi.core import resource
from ooi.occi.infrastructure import compute
from ooi.occi.infrastructure import network
from ooi.occi.infrastructure import network_link
from ooi.occi.infrastructure import storage
from ooi.occi.infrastructure import storage_link
from ooi.occi.infrastructure import templates as infra_templates
from ooi.openstack import contextualization
from ooi.openstack import network as os_network
from ooi.openstack import templates
from ooi.tests import base
from ooi.tests import fakes


class TestQueryController(base.TestController):
    def setUp(self):
        super(TestQueryController, self).setUp()
        self.controller = query.Controller(mock.MagicMock(), None)

    @mock.patch.object(query.Controller, "_os_tpls")
    @mock.patch.object(query.Controller, "_resource_tpls")
    @mock.patch.object(query.Controller, "_ip_pools")
    def test_index(self, m_res, m_os, m_pools):
        tenant = fakes.tenants["foo"]
        req = self._build_req(tenant["id"])

        f = fakes.flavors[1]
        res_tpl = templates.OpenStackResourceTemplate(f["id"], f["name"],
                                                      f["vcpus"], f["ram"],
                                                      f["disk"])
        m_res.return_value = [res_tpl]
        i = fakes.images["foo"]
        os_tpl = templates.OpenStackOSTemplate(i["id"], i["name"])
        m_os.return_value = [os_tpl]
        ip_pool = os_network.OSFloatingIPPool("foo")
        m_pools.return_value = [ip_pool]

        expected = [
            res_tpl,
            os_tpl,
            ip_pool,
            # OCCI Core Kinds:
            entity.Entity.kind,
            resource.Resource.kind,
            link.Link.kind,

            # OCCI infra Compute:
            compute.ComputeResource.kind,
            compute.start,
            compute.stop,
            compute.restart,
            compute.suspend,

            # OCCI infra Storage
            storage.StorageResource.kind,
            storage_link.StorageLink.kind,
            storage.online,
            storage.offline,
            storage.backup,
            storage.snapshot,
            storage.resize,

            # OCCI infra network
            network.NetworkResource.kind,
            network.up,
            network.down,
            network.ip_network,
            network_link.NetworkInterface.kind,
            network_link.ip_network_interface,

            # OCCI infra compute mixins
            infra_templates.os_tpl,
            infra_templates.resource_tpl,

            # OpenStack Contextualization
            contextualization.user_data,
            contextualization.public_key,
        ]

        ret = self.controller.index(req)
        self.assertItemsEqual(expected, ret)

    @mock.patch.object(query.Controller, "_os_tpls")
    @mock.patch.object(query.Controller, "_resource_tpls")
    @mock.patch.object(query.Controller, "_ip_pools")
    def test_index_neutron(self, m_res, m_os, m_pools):
        neutron_controller = query.Controller(mock.MagicMock(),
                                              None, "http://foo")
        tenant = fakes.tenants["foo"]
        req = self._build_req(tenant["id"])

        f = fakes.flavors[1]
        res_tpl = templates.OpenStackResourceTemplate(f["id"], f["name"],
                                                      f["vcpus"], f["ram"],
                                                      f["disk"])
        m_res.return_value = [res_tpl]
        i = fakes.images["foo"]
        os_tpl = templates.OpenStackOSTemplate(i["id"], i["name"])
        m_os.return_value = [os_tpl]
        ip_pool = os_network.OSFloatingIPPool("foo")
        m_pools.return_value = [ip_pool]

        expected = [
            res_tpl,
            os_tpl,
            ip_pool,
            # OCCI Core Kinds:
            entity.Entity.kind,
            resource.Resource.kind,
            link.Link.kind,

            # OCCI infra Compute:
            compute.ComputeResource.kind,
            compute.start,
            compute.stop,
            compute.restart,
            compute.suspend,

            # OCCI infra Storage
            storage.StorageResource.kind,
            storage_link.StorageLink.kind,
            storage.online,
            storage.offline,
            storage.backup,
            storage.snapshot,
            storage.resize,

            # OCCI infra network
            network.NetworkResource.kind,
            network.up,
            network.down,
            os_network.neutron_network,
            network.ip_network,
            network_link.NetworkInterface.kind,
            network_link.ip_network_interface,

            # OCCI infra compute mixins
            infra_templates.os_tpl,
            infra_templates.resource_tpl,

            # OpenStack Contextualization
            contextualization.user_data,
            contextualization.public_key,
        ]

        ret = neutron_controller.index(req)
        self.assertItemsEqual(expected, ret)

    @mock.patch.object(helpers.OpenStackHelper, "get_flavors")
    def test_get_resource_tpls(self, m_get_flavors):
        m_get_flavors.return_value = fakes.flavors.values()
        ret = self.controller._resource_tpls(None)
        expected = []
        for f in fakes.flavors.values():
            expected.append(
                templates.OpenStackResourceTemplate(f["id"], f["name"],
                                                    f["vcpus"], f["ram"],
                                                    f["disk"])
            )
        # FIXME(aloga): this won't work until we create the correct equality
        # functions
        # self.assertItemsEqual(expected, ret)
        for i in range(len(expected)):
            self.assertEqual(expected[i].title, ret[i].title)

    @mock.patch.object(helpers.OpenStackHelper, "get_flavors")
    def test_get_resource_tpls_empty(self, m_get_flavors):
        m_get_flavors.return_value = []
        ret = self.controller._resource_tpls(None)
        expected = []
        self.assertEqual(expected, ret)

    @mock.patch.object(helpers.OpenStackHelper, "get_images")
    def test_get_os_tpls(self, m_get_images):
        m_get_images.return_value = fakes.images.values()
        ret = self.controller._os_tpls(None)
        expected = []
        for i in fakes.images.values():
            expected.append(
                templates.OpenStackOSTemplate(i["id"], i["name"]),
            )
        # FIXME(aloga): this won't work until we create the correct equality
        # functions
        # self.assertItemsEqual(expected, ret)
        for i in range(len(expected)):
            self.assertEqual(expected[i].title, ret[i].title)

    @mock.patch.object(helpers.OpenStackHelper, "get_images")
    def test_get_os_tpls_empty(self, m_get_images):
        m_get_images.return_value = []
        ret = self.controller._os_tpls(None)
        expected = []
        self.assertEqual(expected, ret)

    @mock.patch.object(helpers.OpenStackHelper, "get_floating_ip_pools")
    def test_get_ip_pools(self, m_get_pools):
        for tenant in fakes.tenants.values():
            pools = fakes.pools[tenant["id"]]
            m_get_pools.return_value = pools
            expected = [os_network.OSFloatingIPPool(p['id']) for p in pools]
            ret = self.controller._ip_pools(None)
            # FIXME(aloga): this won't work til we create the correct equality
            # functions
            # self.assertItemsEqual(expected, ret)
            for i in range(len(expected)):
                self.assertEqual(expected[i].title, ret[i].title)

    @mock.patch.object(helpers.OpenStackHelper, "get_floating_ip_pools")
    def test_get_ip_pools_empty(self, m_get_pools):
        m_get_pools.return_value = []
        ret = self.controller._ip_pools(None)
        expected = []
        self.assertEqual(expected, ret)



# -*- coding: utf-8 -*-

# Copyright 2015 LIP - INDIGO-DataCloud
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json
import uuid

import mock

from ooi.api import helpers_neutron
from ooi import exception
from ooi.tests import base
from ooi.tests import fakes_network as fakes
from ooi import utils


class TestNetOpenStackHelper(base.TestCase):
    def setUp(self):
        super(TestNetOpenStackHelper, self).setUp()
        self.version = "version foo bar baz"
        self.helper = helpers_neutron.OpenStackNeutron(self.version)
        self.translation = {"networks": {"occi.core.title": "name"
                                         },
                            "subnets": {"occi.core.id": "network_id",
                                        "org.openstack.network.ip_version":
                                            "ip_version",
                                        "occi.network.address": "cidr",
                                        "occi.network.gateway":
                                            "gateway_ip"
                                        }
                            }

    @mock.patch.object(helpers_neutron.OpenStackNeutron, "_make_get_request")
    def test_index(self, m):
        resp = fakes.create_fake_json_resp({"networks": ["FOO"]}, 200)
        req_mock = mock.MagicMock()
        req_mock.get_response.return_value = resp
        m.return_value = req_mock
        ret = self.helper.list_resources(None, 'networks', None)
        self.assertEqual(["FOO"], ret)
        m.assert_called_with(None, "/networks", None)

    @mock.patch.object(helpers_neutron.OpenStackNeutron, "_get_req")
    def test_get_list_with_public(self, m):
        id_public = uuid.uuid4().hex
        id_private = uuid.uuid4().hex
        net_private = {"status": "ACTIVE", "id": id_private}
        net_public = {"status": "ACTIVE", "id": id_public,
                      'router:external': True}
        resp = fakes.create_fake_json_resp({
            "networks": [net_private, net_public]}, 200)
        req_mock = mock.MagicMock()
        req_mock.get_response.return_value = resp
        m.return_value = req_mock
        ret = self.helper.list_networks(None)
        self.assertEqual(2, ret.__len__())
        self.assertEqual(id_private, ret[0]['id'])
        self.assertEqual('PUBLIC', ret[1]['id'])
        m.assert_called_with(None, method="GET",
                             path="/networks", query_string=None)

    @mock.patch.object(helpers_neutron.OpenStackNeutron, "_get_req")
    def test_index_req(self, m):
        resp = fakes.create_fake_json_resp({"networks": ["FOO"]}, 200)
        req_mock = mock.MagicMock()
        req_mock.get_response.return_value = resp
        m.return_value = req_mock
        ret = self.helper.list_resources(None, 'networks', None)
        self.assertEqual(["FOO"], ret)
        m.assert_called_with(None, method="GET",
                             path="/networks", query_string=None)

    @mock.patch.object(helpers_neutron.OpenStackNeutron, "_make_get_request")
    def test_index3(self, m):
        id = uuid.uuid4().hex
        resp = fakes.create_fake_json_resp({"networks": [{"id": id}]}, 200)
        req_mock = mock.MagicMock()
        req_mock.get_response.return_value = resp
        m.return_value = req_mock
        ret = self.helper.list_networks(None)
        self.assertEqual(id, ret[0]['id'])
        m.assert_called_with(None, "/networks", None)

    @mock.patch.object(helpers_neutron.OpenStackNeutron, "_make_get_request")
    def test_get_network(self, m):
        id = uuid.uuid4().hex
        resp = fakes.create_fake_json_resp(
            {"network": {"status": "ACTIVE", "id": id}}, 200)
        req_mock = mock.MagicMock()
        req_mock.get_response.return_value = resp
        m.return_value = req_mock
        ret = self.helper.get_network_details(None, id)
        self.assertEqual("active", ret["state"])
        m.assert_called_with(None, "/networks/%s" % id)

    @mock.patch.object(helpers_neutron.OpenStackNeutron, "_get_req")
    def test_get_network_public(self, m):
        id = 'PUBLIC'
        resp_get_public = fakes.create_fake_json_resp(
            {"networks": [{"status": "ACTIVE", "id": id}]}, 200)
        req_mock_public = mock.MagicMock()
        req_mock_public.get_response.return_value = resp_get_public
        resp = fakes.create_fake_json_resp(
            {"network": {"status": "ACTIVE", "id": id}}, 200)
        req_mock = mock.MagicMock()
        req_mock.get_response.return_value = resp
        m.side_effect = [req_mock_public, req_mock]
        ret = self.helper.get_network_details(None, id)
        self.assertEqual("active", ret["state"])
        m.assert_called_with(None, method="GET",
                             path="/networks/%s" % id,
                             query_string=None)

    @mock.patch.object(helpers_neutron.OpenStackNeutron, "_get_req")
    def test_get_network_req(self, m):
        id = uuid.uuid4().hex
        resp = fakes.create_fake_json_resp(
            {"network": {"status": "ACTIVE", "id": id}}, 200)
        req_mock = mock.MagicMock()
        req_mock.get_response.return_value = resp
        m.return_value = req_mock
        ret = self.helper.get_network_details(None, id)
        self.assertEqual("active", ret["state"])
        m.assert_called_with(None, method="GET",
                             path="/networks/%s" % id,
                             query_string=None)

    @mock.patch.object(helpers_neutron.OpenStackNeutron, "_make_get_request")
    def test_get_subnetwork(self, m):
        id = uuid.uuid4().hex
        resp = fakes.create_fake_json_resp(
            {"subnet": ["FOO"]}, 200)
        req_mock = mock.MagicMock()
        req_mock.get_response.return_value = resp
        m.return_value = req_mock
        ret = self.helper.get_resource(None, 'subnets', id)
        self.assertEqual(["FOO"], ret)
        m.assert_called_with(None, "/subnets/%s" % id)

    @mock.patch.object(helpers_neutron.OpenStackNeutron, "_make_get_request")
    def test_get_network_with_subnet(self, m):
        id = uuid.uuid4().hex
        address = uuid.uuid4().hex
        resp = fakes.create_fake_json_resp(
            {"network": {"status": "ACTIVE", "id": id, "subnets": [2]},
             "subnet": {"cidr": address}}, 200
        )
        req_mock = mock.MagicMock()
        req_mock.get_response.return_value = resp
        m.return_value = req_mock
        ret = self.helper.get_network_details(None, id)
        self.assertEqual("active", ret["state"])
        self.assertEqual(address, ret["address"])
        m.assert_called_with(req_mock, "/subnets/2")

    @mock.patch.object(helpers_neutron.OpenStackNeutron, "_get_req")
    def test_get_network_with_subnet2(self, m):
        id = uuid.uuid4().hex
        address = '90349034'
        resp = fakes.create_fake_json_resp(
            {"network": {"status": "DOWN", "id": id, "subnets": [2]},
             "subnet": {"cidr": address}}, 200
        )
        req_mock = mock.MagicMock()
        req_mock.get_response.return_value = resp
        m.return_value = req_mock
        ret = self.helper.get_network_details(None, id)
        self.assertEqual("inactive", ret["state"])
        self.assertEqual(address, ret["address"])
        m.assert_called_with(req_mock, method="GET",
                             path="/subnets/2", query_string=None)

    @mock.patch.object(helpers_neutron.OpenStackNeutron,
                       "_make_create_request")
    def test_create_only_network(self, m):
        name = "name_net"
        net_id = uuid.uuid4().hex
        state = "ACTIVE"
        project = "project_id"
        parameters = {"occi.core.title": name,
                      "occi.core.id": net_id,
                      "occi.network.state": state,
                      "project": project,
                      }
        resp = fakes.create_fake_json_resp(
            {"network": {"network_id": net_id}}, 201)
        req_mock = mock.MagicMock()
        req_mock.get_response.return_value = resp
        m.return_value = req_mock
        ret = self.helper.create_resource(None, 'networks', parameters)
        self.assertEqual(net_id, ret["network_id"])
        m.assert_called_with(None, "networks", parameters)

    @mock.patch.object(helpers_neutron.OpenStackNeutron,
                       "_make_create_request")
    def test_create_resource_net_subnet(self, m):
        name = "name_net"
        net_id = uuid.uuid4().hex
        state = "ACTIVE"
        project = "project_id"
        ip_version = 4
        cidr = "0.0.0.0"
        gate_way = "0.0.0.1"
        subnet_id = uuid.uuid4().hex
        parameters = {"occi.core.title": name,
                      "occi.core.id": net_id,
                      "occi.network.state": state,
                      "X_PROJECT_ID": project,
                      "occi.network.ip_version": ip_version,
                      "occi.networkinterface.address": cidr,
                      "occi.networkinterface.gateway": gate_way
                      }
        resp = fakes.create_fake_json_resp(
            {"network": {"id": net_id},
             "subnet": {"id": subnet_id}}, 201
        )
        req_mock = mock.MagicMock()
        req_mock.get_response.return_value = resp
        m.return_value = req_mock
        ret = self.helper.create_resource(None, 'networks', parameters)
        self.assertEqual(net_id, ret["id"])
        ret2 = self.helper.create_resource(None, 'subnets', parameters)
        self.assertEqual(subnet_id, ret2["id"])
        m.assert_called_with(None, "subnets", parameters)

    @mock.patch.object(helpers_neutron.OpenStackNeutron, "_get_req")
    def test_create_resource_net_subnet_req(self, m):
        name = "name_net"
        net_id = uuid.uuid4().hex
        state = "ACTIVE"
        project = "project_id"
        ip_version = uuid.uuid4().hex
        cidr = "0.0.0.0"
        gate_way = "0.0.0.1"
        subnet_id = uuid.uuid4().hex
        parameters = {"occi.core.title": name,
                      "occi.core.id": net_id,
                      "occi.network.state": state,
                      "X_PROJECT_ID": project,
                      "org.openstack.network.ip_version": ip_version,
                      "occi.network.address": cidr,
                      "occi.network.gateway": gate_way
                      }
        resp = fakes.create_fake_json_resp(
            {"network": {"id": net_id},
             "subnet": {"id": subnet_id}}, 201
        )
        req_mock = mock.MagicMock()
        req_mock.get_response.return_value = resp
        m.return_value = req_mock
        ret = self.helper.create_resource(None, 'networks', parameters)
        self.assertEqual(net_id, ret["id"])
        ret2 = self.helper.create_resource(None, 'subnets', parameters)
        self.assertEqual(subnet_id, ret2["id"])
        m.assert_called_with(None,
                             path="/subnets",
                             content_type="application/json",
                             body=json.dumps(utils.make_body(
                                 "subnet", parameters)),
                             method="POST")

    @mock.patch.object(helpers_neutron.OpenStackNeutron, "_make_put_request")
    def test_add_router_interface(self, m):
        router_id = uuid.uuid4().hex
        subnet_id = uuid.uuid4().hex
        port_id = uuid.uuid4().hex
        resp = fakes.create_fake_json_resp(
            {"port_id": port_id, "subnet_id": subnet_id}, 201
        )
        req_mock = mock.MagicMock()
        req_mock.get_response.return_value = resp
        m.return_value = req_mock
        ret = self.helper._add_router_interface(None, router_id, subnet_id)
        self.assertEqual(port_id, ret["port_id"])
        self.assertEqual(subnet_id, ret["subnet_id"])
        path = "/routers/%s/add_router_interface" % router_id
        param = {"subnet_id": subnet_id}
        m.assert_called_with(None, path, param)

    @mock.patch.object(helpers_neutron.OpenStackNeutron, "_make_put_request")
    def test_remove_router_interface(self, m):
        router_id = uuid.uuid4().hex
        subnet_id = uuid.uuid4().hex
        port_id = uuid.uuid4().hex
        resp = fakes.create_fake_json_resp(
            {"port_id": port_id, "subnet_id": subnet_id}, 201
        )
        req_mock = mock.MagicMock()
        req_mock.get_response.return_value = resp
        m.return_value = req_mock
        ret = self.helper._remove_router_interface(None, router_id, port_id)
        self.assertEqual(port_id, ret["port_id"])
        self.assertEqual(subnet_id, ret["subnet_id"])
        path = "/routers/%s/remove_router_interface" % router_id
        param = {"port_id": port_id}
        m.assert_called_with(None, path, param)

    @mock.patch.object(helpers_neutron.OpenStackNeutron, "list_resources")
    def test_get_public_network(self, m):
        public_id = uuid.uuid4().hex
        m.return_value = [{"id": public_id}]
        ret = self.helper._get_public_network(None)
        att_public = {"router:external": True}
        self.assertEqual(public_id, ret)
        m.assert_called_with(None, 'networks', att_public)

    @mock.patch.object(helpers_neutron.OpenStackNeutron, "create_resource")
    def test_add_floating_ip(self, m):
        port_id = uuid.uuid4().hex
        public_net = uuid.uuid4().hex
        f_ip = uuid.uuid4().hex
        ip = '0.0.0.1'
        m.return_value = {"id": f_ip,
                          'floating_ip_address': ip}
        ret = self.helper._add_floating_ip(None, public_net, port_id)
        attributes_port = {
            "floating_network_id": public_net,
            "port_id": port_id
        }
        self.assertEqual(f_ip, ret['id'])
        self.assertEqual(ip, ret['floating_ip_address'])
        m.assert_called_with(None, 'floatingips', attributes_port)

    @mock.patch.object(helpers_neutron.OpenStackNeutron, "delete_resource")
    @mock.patch.object(helpers_neutron.OpenStackNeutron, "list_resources")
    def test_remove_floating_ip(self, m_list, m_del):
        ip = '1.0.0.0'
        public_net = uuid.uuid4().hex
        f_ip = uuid.uuid4().hex
        m_list.return_value = [{'id': f_ip}]
        m_del.return_value = []
        ret = self.helper._remove_floating_ip(None, public_net, ip)
        attributes_port = {
            "floating_network_id": public_net,
            "floating_ip_address": ip
        }
        self.assertEqual([], ret)
        m_list.assert_called_with(None, 'floatingips', attributes_port)
        m_del.assert_called_with(None, 'floatingips', f_ip)

    @mock.patch.object(helpers_neutron.OpenStackNeutron, "create_resource")
    @mock.patch.object(helpers_neutron.OpenStackNeutron, "list_resources")
    @mock.patch.object(helpers_neutron.OpenStackNeutron,
                       "_add_router_interface")
    def test_create_full_network(self, add_if, list_net, cre_net):
        name = "name_net"
        net_id = uuid.uuid4().hex
        subnet_id = uuid.uuid4().hex
        router_id = uuid.uuid4().hex
        public_net = uuid.uuid4().hex
        state = "ACTIVE"
        ip_version = 4
        cidr = "0.0.0.0/24"
        gate_way = "0.0.0.1"
        parameters = {"occi.core.title": name,
                      "occi.core.id": net_id,
                      "occi.network.state": state,

                      "org.openstack.network.ip_version": ip_version,
                      "occi.network.address": cidr,
                      "occi.network.gateway": gate_way
                      }
        cre_net.side_effect = [{'id': net_id,
                                "status": 'active',
                                "name": 'xx'},
                               {"id": subnet_id,
                                "cidr": cidr,
                                "gateway_ip": gate_way,
                                },
                               {"id": router_id},
                               {"id": 0}]
        list_net.return_value = [{'id': public_net}]
        ret = self.helper.create_network(None,
                                         name=name,
                                         cidr=cidr,
                                         gateway=gate_way,
                                         ip_version=ip_version)
        self.assertEqual(net_id, ret["id"])
        param = utils.translate_parameters(
            self.translation["networks"], parameters)
        self.assertEqual((None, 'networks',
                          param),
                         cre_net.call_args_list[0][0])
        param_subnet = utils.translate_parameters(
            self.translation["subnets"], parameters)
        param_subnet['network_id'] = net_id
        self.assertEqual((None, 'subnets',
                          param_subnet),
                         cre_net.call_args_list[1][0])
        self.assertEqual((None, 'routers',
                          {'external_gateway_info': {
                              'network_id': public_net
                          }
                          }),
                         cre_net.call_args_list[2][0])
        add_if.assert_called_with(None, router_id, subnet_id)
        self.assertEqual(cidr, ret['address'])
        self.assertEqual(gate_way, ret['gateway'])

    @mock.patch.object(helpers_neutron.OpenStackNeutron, "delete_resource")
    @mock.patch.object(helpers_neutron.OpenStackNeutron, "list_resources")
    @mock.patch.object(helpers_neutron.OpenStackNeutron,
                       "_remove_router_interface")
    def test_delete_network(self, m_if, m_list, m_del):
        net_id = uuid.uuid4().hex
        router_id = uuid.uuid4().hex
        m_del.side_effect = [{0},
                             {0},
                             {0},
                             ]
        port1 = {'id': 1,
                 'device_owner': 'network:router_interface',
                 'device_id': router_id
                 }
        port2 = {'id': 2, 'device_owner': 'nova'}
        m_list.return_value = [port1, port2]
        m_del.side_effect = [{0}, {0}, []]
        m_if.return_value = []
        ret = self.helper.delete_network(None, net_id)
        self.assertEqual(ret, [])
        self.assertEqual((None, 'routers',
                          port1['device_id']),
                         m_del.call_args_list[0][0])
        self.assertEqual((None, 'ports',
                          port2['id']),
                         m_del.call_args_list[1][0])
        self.assertEqual((None, 'networks',
                          net_id),
                         m_del.call_args_list[2][0])

    @mock.patch.object(helpers_neutron.OpenStackNeutron,
                       "_make_delete_request")
    def test_delete_network_resource_make_mock(self, m):
        resp = fakes.create_fake_json_resp({"network": []}, 204)
        req_mock = mock.MagicMock()
        req_mock.get_response.return_value = resp
        m.return_value = req_mock
        ret = self.helper.delete_resource(None, 'networks', 1)
        self.assertEqual(ret, [])
        m.assert_called_with(None, "/networks", 1)

    @mock.patch.object(helpers_neutron.OpenStackNeutron, "_get_req")
    def test_response_delete_network_resource(self, m):
        resp = fakes.create_fake_json_resp({"network": []}, 204)
        req_mock = mock.MagicMock()
        req_mock.get_response.return_value = resp
        m.return_value = req_mock
        id = 1
        ret = self.helper.delete_resource(None, 'networks', id)
        self.assertEqual(ret, [])
        m.assert_called_with(None, method="DELETE",
                             path="/networks/1")

    @mock.patch.object(helpers_neutron.OpenStackNeutron, "_get_public_network")
    @mock.patch.object(helpers_neutron.OpenStackNeutron, "list_resources")
    @mock.patch.object(helpers_neutron.OpenStackNeutron, "_add_floating_ip")
    def test_assign_floating_ip(self, m_add, m_list, m_get_net):
        compute_id = uuid.uuid4().hex
        net_id = uuid.uuid4().hex
        f_id = uuid.uuid4().hex
        ip = '0.0.0.1'
        port = {'id': 11, 'network_id': net_id,
                'device_owner': 'nova'}
        param = {'device_id': compute_id}
        m_get_net.return_value = net_id
        m_list.return_value = [port]
        m_add.return_value = {"id": f_id,
                              'floating_ip_address': ip,
                              'floating_network_id': '84'}
        ret = self.helper.assign_floating_ip(None, device_id=compute_id)
        self.assertEqual(net_id, ret['network_id'])
        self.assertEqual(ip, ret['ip'])
        m_list.assert_called_with(None, 'ports', param)
        m_add.assert_called_with(None, net_id, port['id'])

    @mock.patch.object(helpers_neutron.OpenStackNeutron, "_get_public_network")
    @mock.patch.object(helpers_neutron.OpenStackNeutron, "_remove_floating_ip")
    def test_release_floating_ip(self, m_add, m_get_net):
        ip = '22.0.0.1'
        net_id = 'PUBLIC'
        iface = {'net_id': net_id,
                 'ip': ip}
        m_get_net.return_value = net_id
        m_add.return_value = []
        ret = self.helper.release_floating_ip(None, iface)
        self.assertEqual([], ret)
        m_add.assert_called_with(None, net_id, ip)

    @mock.patch.object(helpers_neutron.OpenStackNeutron, "create_resource")
    def test_create_port(self, m_create):
        ip = '22.0.0.1'
        net_id = uuid.uuid4().hex
        mac = '890234'
        device_id = uuid.uuid4().hex
        p = {"network_id": net_id, 'device_id': device_id,
             "fixed_ips": [{"ip_address": ip}],
             "mac_address": mac, "status": "ACTIVE"
             }
        m_create.return_value = p
        ret = self.helper.create_port(None, net_id, device_id)
        self.assertEqual(device_id, ret['compute_id'])
        self.assertEqual(ip, ret['ip'])
        self.assertEqual(net_id, ret['network_id'])
        self.assertEqual(mac, ret['mac'])

    @mock.patch.object(helpers_neutron.OpenStackNeutron, "list_resources")
    @mock.patch.object(helpers_neutron.OpenStackNeutron, "delete_resource")
    def test_delete_port(self, m_delete, m_list):
        port_id = uuid.uuid4().hex
        p = [{'id': port_id}]
        m_list.return_value = p
        m_delete.return_value = []
        iface = {'compute_id': None,
                 'mac': None}
        ret = self.helper.delete_port(None, iface)
        self.assertEqual([], ret)

    @mock.patch.object(helpers_neutron.OpenStackNeutron, "list_resources")
    def test_list_port_not_found(self, m_list):
        iface = {'compute_id': None,
                 'mac': None}
        m_list.return_value = []
        self.assertRaises(exception.LinkNotFound,
                          self.helper.delete_port,
                          None,
                          iface)



# -*- coding: utf-8 -*-

# Copyright 2015 Spanish National Research Council
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from ooi.occi.core import action
from ooi.occi.core import category
from ooi.occi import helpers


class Kind(category.Category):
    """OCCI Kind.

    The Kind type is the core of the type classification system built into the
    OCCI Core Model. Kind is a specialisation of Category and introduces
    additional resource capabilities in terms of Actions.
    """

    def __init__(self, scheme, term, title, attributes=None, location=None,
                 related=[], actions=[]):
        super(Kind, self).__init__(scheme, term, title, attributes=attributes,
                                   location=location)

        helpers.check_type(related, Kind)
        helpers.check_type(actions, action.Action)

        self.related = related
        self.actions = actions

    def _class_name(self):
        return "kind"



# -*- coding: utf-8 -*-

# Copyright 2015 Spanish National Research Council
# Copyright 2015 LIP - INDIGO-DataCloud
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from ooi.occi.core import attribute as attr
from ooi.occi.core import mixin
from ooi.occi.infrastructure import network
from ooi.occi.infrastructure import network_link
from ooi.openstack import helpers


class OSFloatingIPPool(mixin.Mixin):
    scheme = helpers.build_scheme("network/floatingippool")

    def __init__(self, pool=None):
        super(OSFloatingIPPool, self).__init__(self.scheme, pool, pool)


class OSNetworkInterface(network_link.NetworkInterface):
    attributes = attr.AttributeCollection(["occi.networkinterface.address",
                                           "occi.networkinterface.gateway",
                                           "occi.networkinterface.allocation"])

    def __init__(self, source, target, mac, address, ip_id=None,
                 pool=None, state='active'):
        link_id = '_'.join([source.id, target.id, address])
        mixins = [network_link.ip_network_interface]
        if pool:
            mixins.append(OSFloatingIPPool(pool))
        super(OSNetworkInterface, self).__init__(mixins, source, target,
                                                 link_id, "eth0", mac,
                                                 state)
        self.ip_id = ip_id
        self.attributes["occi.networkinterface.address"] = (
            attr.MutableAttribute("occi.networkinterface.address", address))
        self.attributes["occi.networkinterface.gateway"] = (
            attr.MutableAttribute("occi.networkinterface.gateway", None))
        self.attributes["occi.networkinterface.allocation"] = (
            attr.MutableAttribute("occi.networkinterface.allocation",
                                  "dynamic"))

    @property
    def address(self):
        return self.attributes["occi.networkinterface.address"].value

    @address.setter
    def address(self, value):
        self.attributes["occi.networkinterface.address"].value = value

    @property
    def gateway(self):
        return self.attributes["occi.networkinterface.gateway"].value

    @gateway.setter
    def gateway(self, value):
        self.attributes["occi.networkinterface.gateway"].value = value

    @property
    def allocation(self):
        return self.attributes["occi.networkinterface.allocation"].value

    @allocation.setter
    def allocation(self, value):
        self.attributes["occi.networkinterface.allocation"].value = value


class OSNetwork(mixin.Mixin):
    scheme = helpers.build_scheme("infrastructure/network")

    def __init__(self, pool=None):
        term = "osnetwork"
        title = "openstack network"

        super(OSNetwork, self).__init__(
            scheme=self.scheme,
            term=term,
            title=title,
            attributes=attr.AttributeCollection([
                "org.openstack.network.ip_version"
            ])
        )


os_network = OSNetwork()


class OSNetworkResource(network.NetworkResource):

    attributes = attr.AttributeCollection([
        "org.openstack.network.ip_version",
        "occi.network.address",
        "occi.network.gateway",
        "occi.network.allocation",
    ])

    def __init__(self, title=None, summary=None,
                 id=None, vlan=None, label=None, state=None,
                 address=None, gateway=None, ip_version=None, allocation=None):

        super(OSNetworkResource,
              self).__init__(title=title,
                             summary=summary, id=id, vlan=vlan,
                             label=label, state=state,
                             mixins=[network.ip_network, OSNetwork()])
        # subnet
        self.attributes["org.openstack.network.ip_version"] = (
            attr.MutableAttribute(
                "org.openstack.network.ip_version", ip_version))
        self.attributes["occi.network.address"] = (
            attr.MutableAttribute(
                "occi.network.address", address))
        self.attributes["occi.network.gateway"] = (
            attr.MutableAttribute(
                "occi.network.gateway", gateway))
        self.attributes["occi.network.allocation"] = (
            attr.MutableAttribute(
                "occi.network.allocation", allocation))

    @property
    def ip_version(self):
        return self.attributes["org.openstack.network.ip_version"].value

    @ip_version.setter
    def ip_version(self, value):
        self.attributes["org.openstack.network.ip_version"].value = value

    @property
    def address(self):
        return self.attributes["occi.network.address"].value

    @address.setter
    def address(self, value):
        self.attributes["occi.network.address"].value = value

    @property
    def gateway(self):
        return self.attributes["occi.network.gateway"].value

    @gateway.setter
    def gateway(self, value):
        self.attributes["occi.network.gateway"].value = value

    @property
    def allocation(self):
        return self.attributes["occi.network.allocation"].value

    @allocation.setter
    def allocation(self, value):
        self.attributes["occi.network.allocation"].value = value


neutron_network = mixin.Mixin(helpers.build_scheme("infrastructure/network"),
                              "neutron", "Network component",
                              )



# -*- coding: utf-8 -*-

# Copyright 2015 Spanish National Research Council
# Copyright 2015 LIP - INDIGO-DataCloud
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from ooi.api import base
import ooi.api.helpers
from ooi.occi.core import entity
from ooi.occi.core import link
from ooi.occi.core import resource
from ooi.occi.infrastructure import compute
from ooi.occi.infrastructure import network
from ooi.occi.infrastructure import network_link
from ooi.occi.infrastructure import storage
from ooi.occi.infrastructure import storage_link
from ooi.occi.infrastructure import templates as infra_templates
from ooi.openstack import contextualization
from ooi.openstack import network as os_network
from ooi.openstack import templates


class Controller(base.Controller):
    def __init__(self, app, openstack_version,
                 neutron_ooi_endpoint=None):
        self.neutron_ooi_endpoint = neutron_ooi_endpoint
        super(Controller, self).__init__(
            app=app,
            openstack_version=openstack_version)
        self.os_helper = ooi.api.helpers.OpenStackHelper(
            self.app,
            self.openstack_version
        )

    def _resource_tpls(self, req):
        flavors = self.os_helper.get_flavors(req)
        occi_resource_templates = []
        if flavors:
            for f in flavors:
                tpl = templates.OpenStackResourceTemplate(f["id"],
                                                          f["name"],
                                                          f["vcpus"],
                                                          f["ram"],
                                                          f["disk"])
                occi_resource_templates.append(tpl)
        return occi_resource_templates

    def _os_tpls(self, req):
        images = self.os_helper.get_images(req)
        occi_os_templates = []
        if images:
            for i in images:
                tpl = templates.OpenStackOSTemplate(i["id"], i["name"])
                occi_os_templates.append(tpl)
        return occi_os_templates

    def _ip_pools(self, req):
        pools = self.os_helper.get_floating_ip_pools(req)
        occi_ip_pools = []
        if pools:
            for p in pools:
                occi_ip_pools.append(os_network.OSFloatingIPPool(p["name"]))
        return occi_ip_pools

    def index(self, req):
        l = []
        # OCCI Core Kinds:
        l.append(entity.Entity.kind)
        l.append(resource.Resource.kind)
        l.append(link.Link.kind)

        # OCCI infra Compute:
        l.append(compute.ComputeResource.kind)
        l.extend(compute.ComputeResource.actions)

        # OCCI infra Storage
        l.append(storage.StorageResource.kind)
        l.append(storage_link.StorageLink.kind)
        l.extend(storage.StorageResource.actions)

        # OCCI infra network
        l.append(network.NetworkResource.kind)
        l.extend(network.NetworkResource.actions)
        if self.neutron_ooi_endpoint:
            l.append(os_network.neutron_network)
        l.append(network.ip_network)
        l.append(network_link.NetworkInterface.kind)
        l.append(network_link.ip_network_interface)

        # OCCI infra compute mixins
        l.append(infra_templates.os_tpl)
        l.append(infra_templates.resource_tpl)

        # OpenStack flavors & images
        l.extend(self._resource_tpls(req))
        l.extend(self._os_tpls(req))

        # OpenStack Contextualization
        l.append(contextualization.user_data)
        l.append(contextualization.public_key)

        # OpenStack Floating IP Pools
        l.extend(self._ip_pools(req))
        return l



# -*- coding: utf-8 -*-

# Copyright 2015 Spanish National Research Council
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from ooi.occi.core import attribute
from ooi.occi.core import mixin
from ooi.openstack import helpers


class OpenStackUserData(mixin.Mixin):
    scheme = helpers.build_scheme("compute/instance")
    term = "user_data"

    def __init__(self, user_data=None):
        attrs = [
            attribute.InmutableAttribute("org.openstack.compute.user_data",
                                         user_data),
        ]

        attrs = attribute.AttributeCollection({a.name: a for a in attrs})

        super(OpenStackUserData, self).__init__(
            OpenStackUserData.scheme, OpenStackUserData.term,
            "Contextualization extension - user_data",
            attributes=attrs)

    @property
    def user_data(self):
        return self.attributes["org.openstack.compute.user_data"].value


class OpenStackPublicKey(mixin.Mixin):
    scheme = helpers.build_scheme("instance/credentials")
    term = "public_key"

    def __init__(self, name=None, data=None):
        attrs = [
            attribute.InmutableAttribute(
                "org.openstack.credentials.publickey.name", name),
            attribute.InmutableAttribute(
                "org.openstack.credentials.publickey.data", data),
        ]

        attrs = attribute.AttributeCollection({a.name: a for a in attrs})

        super(OpenStackPublicKey, self).__init__(
            OpenStackPublicKey.scheme, OpenStackPublicKey.term,
            "Contextualization extension - public_key",
            attributes=attrs)

    @property
    def name(self):
        attr = "org.openstack.credentials.publickey.name"
        return self.attributes[attr].value

    @property
    def data(self):
        attr = "org.openstack.credentials.publickey.data"
        return self.attributes[attr].value


user_data = OpenStackUserData()
public_key = OpenStackPublicKey()



# -*- coding: utf-8 -*-

# Copyright 2015 Spanish National Research Council
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import webob
import webob.dec
import webob.exc

from ooi.tests import base
from ooi.tests import fakes
from ooi import wsgi


class TestMiddleware(base.TestCase):
    """OCCI middleware test without Accept header.

    According to the OCCI HTTP rendering, no Accept header
    means text/plain.
    """

    def setUp(self):
        super(TestMiddleware, self).setUp()

        self.accept = self.content_type = None
        self.application_url = fakes.application_url

        self.occi_string = "OCCI/1.1"

    def get_app(self, resp=None):
        return wsgi.OCCIMiddleware(fakes.FakeApp())

    def assertDefaults(self, result):
        self.assertContentType(result)
        self.assertServerHeader(result)

    def assertContentType(self, result):
        if self.accept in (None, "*/*"):
            expected = "text/plain"
        else:
            expected = self.accept
        self.assertEqual(expected, result.content_type)

    def assertServerHeader(self, result):
        self.assertIn("Server", result.headers)
        self.assertIn(self.occi_string, result.headers["server"])

    def assertExpectedResult(self, expected, result):
        expected = ["%s: %s" % e for e in expected]
        # NOTE(aloga): the order of the result does not matter
        results = result.text.splitlines()
        self.assertItemsEqual(expected, results)

    def assertResultIncludesLink(self, link_id, source, target, result):
        expected_attrs = set([
            'occi.core.source="%s"' % source,
            'occi.core.target="%s"' % target,
            'occi.core.id="%s"' % link_id,
        ])
        for lines in result.text.splitlines():
            r = lines.split(":", 1)
            if r[0] == "Link":
                attrs = set([s.strip() for s in r[1].split(";")])
                if expected_attrs.issubset(attrs):
                    return
        self.fail("Failed to find %s in %s." % (expected_attrs, result))

    def assertResultIncludesLinkAttr(self, link_id, source, target, result):
        expected_attrs = set([
            'occi.core.source="%s"' % source,
            'occi.core.target="%s"' % target,
            'occi.core.id="%s"' % link_id,
        ])
        attrs = set()
        for lines in result.text.splitlines():
            r = lines.split(":", 1)
            if r[0] == "X-OCCI-Attribute":
                attrs.add(r[1].strip())
        if expected_attrs.issubset(attrs):
            return
        self.fail("Failed to find %s in %s." % (expected_attrs, result))

    def _build_req(self, path, tenant_id, **kwargs):
        if self.accept is not None:
            kwargs["accept"] = self.accept

        if self.content_type is not None:
            kwargs["content_type"] = self.content_type

        environ = {"HTTP_X_PROJECT_ID": tenant_id}

        kwargs["base_url"] = self.application_url

        return webob.Request.blank(path, environ=environ, **kwargs)

    def test_404(self):
        result = self._build_req("/", "tenant").get_response(self.get_app())
        self.assertEqual(404, result.status_code)
        self.assertDefaults(result)

    def test_good_user_agent(self):
        req = self._build_req("/", "tenant")
        req.user_agent = "foo OCCI/1.1 bar"
        result = req.get_response(self.get_app())
        self.assertEqual(404, result.status_code)
        self.assertDefaults(result)

    def test_bad_user_agent(self):
        req = self._build_req("/", "tenant")
        req.user_agent = "foo OCCI/2.2 bar"
        result = req.get_response(self.get_app())
        self.assertEqual(501, result.status_code)
        self.assertDefaults(result)

    def test_ugly_user_agent(self):
        req = self._build_req("/", "tenant")
        req.user_agent = "fooOCCI/1.1bar"
        result = req.get_response(self.get_app())
        self.assertEqual(404, result.status_code)
        self.assertDefaults(result)

    def test_400_from_openstack(self):
        @webob.dec.wsgify()
        def _fake_app(req):
            exc = webob.exc.HTTPBadRequest()
            resp = fakes.FakeOpenStackFault(exc)
            return resp

        mdl = wsgi.OCCIMiddleware(_fake_app)
        result = self._build_req("/-/", "tenant").get_response(mdl)
        self.assertEqual(400, result.status_code)
        self.assertDefaults(result)


class TestMiddlewareTextPlain(TestMiddleware):
    """OCCI middleware test with Accept: text/plain."""

    def setUp(self):
        super(TestMiddlewareTextPlain, self).setUp()

        self.accept = "text/plain"

    def test_correct_accept(self):
        self.assertEqual("text/plain", self.accept)


class TestMiddlewareTextOcci(TestMiddleware):
    """OCCI middleware text with Accept: text/occi."""

    def setUp(self):
        super(TestMiddlewareTextOcci, self).setUp()

        self.accept = "text/occi"

    def assertExpectedResult(self, expected, result):
        for hdr, val in expected:
            self.assertIn(val, result.headers.getall(hdr))

    def test_correct_accept(self):
        self.assertEqual("text/occi", self.accept)

    def assertResultIncludesLink(self, link_id, source, target, result):
        expected_attrs = set([
            'occi.core.source="%s"' % source,
            'occi.core.target="%s"' % target,
            'occi.core.id="%s"' % link_id,
        ])
        for val in result.headers.getall("Link"):
            attrs = set([s.strip() for s in val.split(";")])
            if expected_attrs.issubset(attrs):
                return
        self.fail("Failed to find %s in %s." % (expected_attrs, result))

    def assertResultIncludesLinkAttr(self, link_id, source, target, result):
        expected_attrs = set([
            'occi.core.source="%s"' % source,
            'occi.core.target="%s"' % target,
            'occi.core.id="%s"' % link_id,
        ])
        attrs = set([v for v in result.headers.getall("X-OCCI-Attribute")])
        if expected_attrs.issubset(attrs):
            return
        self.fail("Failed to find %s in %s." % (expected_attrs, result))



# -*- coding: utf-8 -*-

# Copyright 2015 Spanish National Research Council
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import webob
import webob.dec
import webob.exc

from ooi import exception
from ooi.tests import base
from ooi import wsgi


class TestResourceExceptionHandler(base.TestCase):
    def test_invalid(self):
        ret = self.return_fault_with_handler(exception.Invalid)
        self.assertIsInstance(ret, wsgi.Fault)
        self.assertIsInstance(ret.wrapped_exc, webob.exc.WSGIHTTPException)
        self.assertEqual(400, ret.wrapped_exc.code)

    def test_not_implemented(self):
        ret = self.return_fault_with_handler(exception.NotImplemented)
        self.assertIsInstance(ret, wsgi.Fault)
        self.assertEqual(webob.exc.HTTPNotImplemented.code,
                         ret.wrapped_exc.code)

    def test_bad_request(self):
        ret = self.return_fault_with_handler(TypeError)
        self.assertIsInstance(ret, wsgi.Fault)
        self.assertIsInstance(ret.wrapped_exc, webob.exc.HTTPBadRequest)

    def test_fault(self):
        fault = wsgi.Fault(webob.exc.HTTPInternalServerError())
        ret = self.return_fault_with_handler(fault)
        self.assertIs(ret, fault)

    def test_http_exception(self):
        ret = self.return_fault_with_handler(webob.exc.HTTPNotFound())
        self.assertIsInstance(ret, wsgi.Fault)
        self.assertIsInstance(ret.wrapped_exc, webob.exc.HTTPNotFound)

    def test_internal_server_error(self):
        ret = self.return_fault_with_handler(Exception)
        self.assertIsInstance(ret, wsgi.Fault)
        self.assertIsInstance(ret.wrapped_exc,
                              webob.exc.HTTPInternalServerError)

    @staticmethod
    def return_fault_with_handler(ex):
        try:
            with wsgi.ResourceExceptionHandler():
                raise ex
        except wsgi.Fault as e:
            return e



# -*- coding: utf-8 -*-

# Copyright 2015 Spanish National Research Council
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import six
import webob.exc

from ooi.occi.core import action
from ooi.occi.core import collection
from ooi.occi.core import entity
from ooi.occi.core import kind
from ooi.occi.core import link
from ooi.occi.core import mixin
from ooi.occi.core import resource
from ooi import utils


class HeaderRenderer(object):
    def __init__(self, obj):
        self.obj = obj

    def render(self, env={}):
        raise NotImplementedError("%s for %s object not implemented" %
                                  (type(self), type(self.obj)))


class ExceptionRenderer(HeaderRenderer):
    def render(self, env={}):
        return [("X-OCCI-Error", self.obj.explanation)]


class CategoryRenderer(HeaderRenderer):
    def _render_location(self, env={}):
        if getattr(self.obj, 'location'):
            url = env.get("application_url", "")
            loc = utils.join_url(url, self.obj.location)
            return ['location="%s"' % loc]
        return []

    def render(self, env={}):
        d = {
            "term": self.obj.term,
            "scheme": self.obj.scheme,
            "class": self.obj.occi_class,
            "title": self.obj.title
        }
        ret = []
        ret.append(('%(term)s; scheme="%(scheme)s"; class="%(class)s"; '
                    'title="%(title)s"') % d)
        for rel in getattr(self.obj, 'related', []):
            d = {"scheme": rel.scheme, "term": rel.term}
            ret.append('rel="%(scheme)s%(term)s"' % d)
        ret.extend(self._render_location(env))
        # FIXME(enolfc): missing attributes and actions
        return [('Category', "; ".join(ret))]


class KindRenderer(CategoryRenderer):
    pass


class ActionRenderer(CategoryRenderer):
    def _render_location(self, env={}):
        """Do not render location for actions."""
        return []

    def render(self, ass_obj=None, env={}):
        # We have an associated object, render it as a link to that object
        if ass_obj is not None:
            url = env.get("application_url", "")
            term = ass_obj.kind.term + "/"
            url = utils.join_url(url, [term, ass_obj.id, self.obj.location])
            d = {"location": url,
                 "rel": self.obj.type_id}
            l = '<%(location)s>; rel="%(rel)s"' % d
            return [('Link', l)]
        else:
            # Otherwise, render as category
            return super(ActionRenderer, self).render(env=env)


class MixinRenderer(CategoryRenderer):
    pass


class CollectionRenderer(HeaderRenderer):
    def render(self, env={}):
        app_url = env.get("application_url", "")
        ret = []
        for what in [self.obj.kinds, self.obj.mixins, self.obj.actions,
                     self.obj.resources, self.obj.links]:
            for el in what:
                url = utils.join_url(app_url, el.location)
                ret.append(('X-OCCI-Location', '%s' % url))
        return ret


class AttributeRenderer(HeaderRenderer):
    def render_attr(self, env={}):
        value_str = ''
        if isinstance(self.obj.value, six.string_types):
            value_str = '"%s"' % self.obj.value
        elif isinstance(self.obj.value, bool):
            value_str = '"%s"' % str(self.obj.value).lower()
        elif isinstance(self.obj.value, entity.Entity):
            app_url = env.get("application_url", "")
            url = utils.join_url(app_url, self.obj.value.location)
            value_str = '"%s"' % url
        else:
            value_str = "%s" % self.obj.value
        return '%s=%s' % (self.obj.name, value_str)

    def render(self, env={}):
        return [('X-OCCI-Attribute', self.render_attr(env))]


class EntityRenderer(HeaderRenderer):
    def render(self, env={}):
        ret = []
        ret.extend(KindRenderer(self.obj.kind).render(env=env))
        for m in self.obj.mixins:
            ret.extend(MixinRenderer(m).render(env=env))
        for a in self.obj.attributes:
            # FIXME(aloga): I dont like this test here
            if self.obj.attributes[a].value is None:
                continue
            r = AttributeRenderer(self.obj.attributes[a])
            ret.extend(r.render(env=env))
        return ret


class LinkRenderer(EntityRenderer):
    def render_link(self, env={}):
        ret = []
        url = env.get("application_url", "")
        url = utils.join_url(url, self.obj.location)
        d = {"location": url,
             "scheme": self.obj.target.kind.scheme,
             "term": self.obj.target.kind.term,
             "self": url}
        l = '<%(location)s>; rel="%(scheme)s%(term)s"; self="%(self)s"' % d
        ret.append(l)
        for a in self.obj.attributes:
            if self.obj.attributes[a].value is None:
                continue
            ret.append(AttributeRenderer(
                self.obj.attributes[a]).render_attr(env=env))
        return [('Link', '; '.join(ret))]


class ResourceRenderer(EntityRenderer):
    def render(self, env={}):
        ret = super(ResourceRenderer, self).render(env)
        for a in self.obj.actions:
            r = ActionRenderer(a)
            ret.extend(r.render(ass_obj=self.obj, env=env))
        for l in self.obj.links:
            ret.extend(LinkRenderer(l).render_link(env=env))
        return ret


_MAP = {
    "action": ActionRenderer,
    "kind": KindRenderer,
    "mixin": MixinRenderer,
    "collection": CollectionRenderer,
    "resource": ResourceRenderer,
    "link": LinkRenderer,
    "exception": ExceptionRenderer,
    None: HeaderRenderer,
}


def get_renderer(obj):
    if isinstance(obj, action.Action):
        type_ = "action"
    elif isinstance(obj, collection.Collection):
        type_ = "collection"
    elif isinstance(obj, mixin.Mixin):
        type_ = "mixin"
    elif isinstance(obj, kind.Kind):
        type_ = "kind"
    elif isinstance(obj, resource.Resource):
        type_ = "resource"
    elif isinstance(obj, link.Link):
        type_ = "link"
    elif isinstance(obj, webob.exc.HTTPException):
        type_ = "exception"
    else:
        type_ = None
    return _MAP.get(type_)(obj)
